
/**
 * DES Cipher Virtual Device Driver
 *
 * Copyright (c) 2010 Tomasz Pieczerak <tphaster at gmail.com>
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/uaccess.h>

#include <linux/device.h>
#include <linux/fs.h>
#include <linux/cdev.h>

#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>

#include <linux/mutex.h>

#include <linux/crypto.h>


MODULE_DESCRIPTION("Driver emulating DES encryption device");
MODULE_AUTHOR("Tomasz Pieczerak (tph)");
MODULE_LICENSE("GPL");

/* Defines */
#define DEVICE_NAME     "des_cipher_vd"
#define MAX_BUF_SIZE    (size_t)(PAGE_SIZE<<8)   /* 1 MiB */

/* Macros */
#define round8(x)       (size_t)((x&7) ? x+8-(x&7) : x)
                        /* round up to the nearest multiple of 8 */

/* Driver Modes */
#define NOOP    0   /* inactive - refuses all read/write operations */
#define REGKEY  1   /* register key mode - write only */
#define ENCRYPT 2   /* encryption mode   - write/read (order important) */
#define DECRYPT 3   /* decryption mode   - write/read (order important) */

/* I/O Control Commands */
#define DESCVD_IOC_MAGIC     'g'    /* a magic 8bit number for ioctl */

#define DESCVD_IOC_NOOP      _IO(DESCVD_IOC_MAGIC, NOOP)
#define DESCVD_IOC_REGKEY    _IO(DESCVD_IOC_MAGIC, REGKEY)
#define DESCVD_IOC_ENCRYPT   _IO(DESCVD_IOC_MAGIC, ENCRYPT)
#define DESCVD_IOC_DECRYPT   _IO(DESCVD_IOC_MAGIC, DECRYPT)

#define DESCVD_IOC_MAXNR     3

/* Prototypes */
static int  __init des_cipher_vd_init ( void );
static void        des_cipher_vd_exit ( void );

static int des_cipher_vd_open    ( struct inode *inode, struct file *filp );
static int des_cipher_vd_release ( struct inode *inode, struct file *filp );

static ssize_t des_cipher_vd_read  ( struct file *filp, char __user *buf,
                                     size_t len, loff_t *offp );
static ssize_t des_cipher_vd_write ( struct file *filp, const char __user *buf,
                                     size_t len, loff_t *offp );
static int     des_cipher_vd_ioctl ( struct inode *inode, struct file *filp,
                                     unsigned int cmd, unsigned long arg );

/* Global Variables */
static        dev_t  des_cipher_vd       = 0;
static struct cdev  *des_cipher_vd_cdev  = NULL;
static struct class *des_cipher_vd_class = NULL;

static struct file_operations des_cipher_vd_fops = {
    .owner   = THIS_MODULE,
    .open    = des_cipher_vd_open,
    .release = des_cipher_vd_release,
    .read    = des_cipher_vd_read,
    .write   = des_cipher_vd_write,
    .ioctl   = des_cipher_vd_ioctl,
};

static struct crypto_cipher *tfm = 0;   /* cipher description */

/* Structures */

/**
 * des_cipher_vd_user - structure describing virtual device state and data
 *                      related to a process, which is currently using it
 *
 * This structure is allocated, when process opens the device and pointer to
 * it is keep in a 'private_data' field of its 'file' structure. Structure
 * is freed, when process closes the device.
 */
typedef struct des_cipher_vd_user {
    int vd_mode;     /* current device mode */
    char key[8];     /* secret key for encryption/decryption (64bit)  */
    char *buffer;    /* buffer for encrypted/decrypted data */
    size_t dat_size; /* size of data in the buffer */
    size_t buf_size; /* real buffer size in the memory (it is data size
                        rounded to pages - as it will be vmalloc'ed */
    struct mutex mutex;     /* mutex for threads' synchronization */
} user_desc;

#define FUSER_DESC  ((user_desc*)(filp->private_data))
                    /* for easier access to it through the 'file' structure */

/**
 * alloc_user_desc - allocate new 'user_desc' structure
 *
 * On success returns a pointer to th newly allocated structure.
 * If there is an error, NULL is returned instead.
 */
static struct des_cipher_vd_user *alloc_user_desc ( void )
{
    user_desc *desc; 

    /**
     * According to http://kerneltrap.org/node/8206 kmalloc'ed memory is
     * unswappable, so it is safe to keep secret key in it.
     */
    if (unlikely((desc = kmalloc(sizeof(user_desc), GFP_KERNEL)) == NULL))
        return NULL;

    /* Consequences of zero'ing the structure:
     *  - mode is set to 'inactive' (NOOP)
     *  - key is zero'ed (HEX: 00 00 00 00 00 00 00 00)
     *  - buffer pointer is set to NULL
     *  - buf_size is set to zero
     *  - dat_size is set to zero
     */
    memset(desc, 0, sizeof(user_desc));

    mutex_init(&desc->mutex);

    return desc;
}

/**
 * free_user_desc - free 'user_desc' structure
 *  @desc: pointer to the structure
 */
static void free_user_desc ( user_desc *desc )
{
    if (desc == NULL)
        return;

    vfree(desc->buffer);

    /**
     * IMPORTANT!!!
     * We have to zero 'user_desc' before freeing it due to privacy reasons
     * (there is a secret key in it), as this memory could be kmalloc'ed again
     * and someone will be able to retrieve information from it.
     */
    memset(desc, 0, sizeof(user_desc));

    kfree(desc);
}


/* Functions */

static int __init des_cipher_vd_init ( void )
{
    int res;

    /* create class describing our virtual device */
    des_cipher_vd_class = class_create(THIS_MODULE, DEVICE_NAME);

    if (IS_ERR(des_cipher_vd_class)) {
        printk(KERN_ERR DEVICE_NAME
               ": error creating device class for udev.\n");
        res = PTR_ERR(des_cipher_vd_class);
        goto error;
    }

    /* dynamically allocate major and minor device numbers */
    res = alloc_chrdev_region(&des_cipher_vd, 0, 1, DEVICE_NAME);

    if (res) {
        printk(KERN_ERR DEVICE_NAME
               ": allocation of the device numbers failed.\n");
        goto error;
    }

    /* allocate structure describing character device */
    des_cipher_vd_cdev = cdev_alloc();
    des_cipher_vd_cdev->ops = &des_cipher_vd_fops;
    des_cipher_vd_cdev->owner = THIS_MODULE;

    /* add a character device to the system */
    res = cdev_add(des_cipher_vd_cdev, des_cipher_vd, 1);

    if (res) {
        printk(KERN_ERR DEVICE_NAME
                ": adding a char device to the system failed.\n");
        goto error;
    }

    /* create a character device and register it with sysfs
     * (there is no need to remember the returned pointer,
     * as we won't create any further sysfs files)
     */
    device_create(des_cipher_vd_class, NULL, des_cipher_vd, NULL, DEVICE_NAME);

    /* allocate structure describing DES cipher */
    tfm = crypto_alloc_cipher("des", 0, 0);
    if (IS_ERR(tfm)) {
        printk(KERN_ERR DEVICE_NAME
               ": cannot allocate structure describing for DES cipher.\n");
        res = PTR_ERR(tfm);
        goto error;
    }

    printk(KERN_INFO DEVICE_NAME ": device creation successful.\n");
    printk(KERN_INFO DEVICE_NAME ": major number is %d, minor is %d.",
                                 MAJOR(des_cipher_vd), MINOR(des_cipher_vd));

    return 0;

error:
    /* free everything we have already allocated */
    des_cipher_vd_exit();
    return res;
}
module_init( des_cipher_vd_init )

static void des_cipher_vd_exit ( void ) {
    /* free structure describing DES cipher */
    if (tfm)
        crypto_free_cipher(tfm);

    /* remove device */
    if (des_cipher_vd && des_cipher_vd_class)
        device_destroy(des_cipher_vd_class, des_cipher_vd);

    /* delete device from the system */
    if (des_cipher_vd_cdev) {
        cdev_del(des_cipher_vd_cdev);
        des_cipher_vd_cdev = NULL;
    }

    /* unregister device major and minor numbers */
    if (des_cipher_vd)
        unregister_chrdev_region(des_cipher_vd, 1);

    /* remove device class */
    if (des_cipher_vd_class) {
        class_destroy(des_cipher_vd_class);
        des_cipher_vd_class = NULL;
    }

    printk(KERN_INFO DEVICE_NAME
                     ": unloading and device removal successful.\n");
}
module_exit( des_cipher_vd_exit )

static int des_cipher_vd_open ( struct inode *inode, struct file *filp )
{
    if (unlikely((filp->private_data = alloc_user_desc()) == NULL))
        return -ENOMEM;

    printk(KERN_INFO DEVICE_NAME ": open successful.\n");

    return 0;
}

static int des_cipher_vd_release ( struct inode *inode, struct file *filp )
{
    free_user_desc(filp->private_data);
    filp->private_data = NULL;
    printk(KERN_INFO DEVICE_NAME ": closed.\n");

    return 0;
}

/**
 * des_cipher_vd_read - read encrypted/decrypted data that was previously
 *                      written to the device
 *
 * See des_cipher_vd_write() for more information.
 */
static ssize_t des_cipher_vd_read  ( struct file *filp, char __user *buf,
                                     size_t len, loff_t *offp )
{
    int ret;

    if(filp->private_data == NULL)
        return -EFAULT;

    if (unlikely(mutex_lock_interruptible(&(FUSER_DESC->mutex))))
        return -EINTR;

    /* read operation is only allowed in encryption or decryption mode */
    if (FUSER_DESC->vd_mode == ENCRYPT || FUSER_DESC->vd_mode == DECRYPT) {
        printk(KERN_INFO DEVICE_NAME ": read buffer (%zu).\n", len);

        if (*offp >= FUSER_DESC->dat_size || len == 0) {
            ret = 0;
            goto done;
        }

        len = min(len, (size_t)(FUSER_DESC->dat_size - *offp));

        if (unlikely(__copy_to_user(buf, (FUSER_DESC->buffer + *offp), len))) {
            ret = -EFAULT;
            goto done;
        }

        /* read data could be zero'ed (so we do it - for safety of data) */
        memset((FUSER_DESC->buffer + *offp), 0, len);

        *offp += len;
        ret = len;
    }
    else
        ret = -EINVAL;

done:
    mutex_unlock(&(FUSER_DESC->mutex));
    return ret;
}

/**
 * des_cipher_vd_write - write data (or secret key) to the device and get it
 *                       encrypted/decrypted (registered)
 *
 * This needs a few words of explanation. Behaviour of write function depends
 * on current device mode (in context of the given process):
 *
 *  - NOOP   : any write operation will be rejected
 *
 *  - REGKEY : you can write your secret 64bit DES key to the device and it
 *             will save and register it; written data must be 64bit (8bytes)
 *             in size; data of any other size will be rejected
 *
 *  - ENCRYPT: write plain text message (up to 1MiB size), it will be encrypted
 *             on-the-fly and you will be able to read it from the device
 *
 *  - DECRYPT: write encrypted message, device will decrypt it on-the-fly and
 *             you will be able to read it
 *
 *  Note#1
 *  Plain text messages will be aligned to blocks of 8 characters (8 bytes),
 *  remaining space will be filled with zeros (\0). Use round8() macro, to
 *  calculate size of message after encryption.
 *  WARNING! This align method may be not safe for DES cipher (knowing it may
 *  make it easier to break the cipher). If you want to prevent this, fill
 *  remaining bytes with some random data.
 *
 *  Note#2
 *  Both decrypted and encrypted messages must be written all at once, because
 *  every successive write operation erase previously written data
 *  (consequently of the on-the-fly encryption/decryption). Processed message
 *  should be read immediately after (but it can be in a few chunks). 
 *
 *  Note#3
 *  Device must be opened in read/write mode (O_RDRW) in order to be useful.
 */
static ssize_t des_cipher_vd_write ( struct file *filp, const char __user *buf,
                                     size_t len, loff_t *offp )
{
    int ret;

    if(filp->private_data == NULL)
        return -EFAULT;

    if (unlikely(mutex_lock_interruptible(&(FUSER_DESC->mutex))))
        return -EINTR;

    if(FUSER_DESC->vd_mode == NOOP) {
        ret = -EINVAL;
        goto done;
    }

    if (FUSER_DESC->vd_mode == REGKEY)
        printk(KERN_INFO DEVICE_NAME ": write key (%zu).\n", len);
    else
        printk(KERN_INFO DEVICE_NAME ": write buffer (%zu).\n", len);

    /* clear buffer - for data safety */
    memset(FUSER_DESC->buffer, 0, FUSER_DESC->dat_size);
    FUSER_DESC->dat_size = 0;
    *offp = 0;

    if (len == 0) {
        ret = 0;
        goto done;
    }
 
    /* save secret 64bit DES key written by user */
    if (FUSER_DESC->vd_mode == REGKEY) {
        if (len != sizeof(FUSER_DESC->key)) {
            ret = -EINVAL;
            goto done;
        }

        if (unlikely(
            __copy_from_user(FUSER_DESC->key, buf, sizeof(FUSER_DESC->key)) ))
        {
            ret = -EFAULT;
            goto done;
        }
    }
    /* preform write operation and encryption/decryption */
    else {
        size_t len8 = round8(len);  /* size aligned to the multiple of 64bit */
        size_t lenpage = PAGE_ALIGN(len8);  /* size aligned to pages */
        size_t clen;

        if (lenpage > MAX_BUF_SIZE) {
            ret = -ENOSPC;
            goto done;
        }

        /* allocate more memory, if it is needed */
        if (lenpage > FUSER_DESC->buf_size) {
            char *tmp;

            if (unlikely((tmp = vmalloc(lenpage)) == NULL)) {
                ret = -ENOMEM;
                goto done;
            }

            printk(KERN_INFO DEVICE_NAME
                             ": allocating new buffer (%zu).\n", lenpage);

            if (FUSER_DESC->buffer != NULL)
                vfree(FUSER_DESC->buffer);

            FUSER_DESC->buffer = tmp;
            FUSER_DESC->buf_size = lenpage;

            memset(FUSER_DESC->buffer, 0, len8);
        }

        if (unlikely(__copy_from_user(FUSER_DESC->buffer, buf, len))) {
            ret = -EFAULT;
            goto done;
        }

        crypto_cipher_setkey(tfm, FUSER_DESC->key, sizeof(FUSER_DESC->key));

        /* encryption mode - encrypt just copied data from user */
        if (FUSER_DESC->vd_mode == ENCRYPT) {
            for (clen = 0; clen < len8; clen += 8)
                crypto_cipher_encrypt_one(tfm, (FUSER_DESC->buffer)+clen,
                                               (FUSER_DESC->buffer)+clen);

            printk(KERN_INFO DEVICE_NAME ": data encrypted\n");
        }
        /* decryption mode - decrypt just copied data from user */
        else if (FUSER_DESC->vd_mode == DECRYPT) {
            for (clen = 0; clen < len8; clen += 8)
                crypto_cipher_decrypt_one(tfm, (FUSER_DESC->buffer)+clen,
                                               (FUSER_DESC->buffer)+clen);

            printk(KERN_INFO DEVICE_NAME ": data decrypted\n");
        }

        /* encrypted/decrypted data is aligned to 64bit blocks, because DES
         * cipher block size; given plain data will be aligned with zeros
         * (0x00) and so decrypted data also contains them (but it is no
         * problem for text messages)
         */
        FUSER_DESC->dat_size = len8;
    }
    ret = len;

done:
    mutex_unlock(&(FUSER_DESC->mutex));
    return ret;
}

/**
 * des_cipher_vd_ioctl - change device current mode
 *
 * For more information see "Device Modes" and "I/O Control Commands" at the
 * beginning of this file and des_cipher_vd_write().
 */
static int des_cipher_vd_ioctl ( struct inode *inode, struct file *filp,
                                 unsigned int cmd, unsigned long arg )
{
    if (_IOC_TYPE(cmd) != DESCVD_IOC_MAGIC)
        return -ENOTTY;
    if (_IOC_NR(cmd) > DESCVD_IOC_MAXNR)
        return -ENOTTY;
    if(filp->private_data == NULL)
        return -EFAULT;

    if (unlikely(mutex_lock_interruptible(&(FUSER_DESC->mutex))))
        return -EINTR;

    /* any mode switching clears the data buffer */
    if (FUSER_DESC->buffer != NULL) {
        memset(FUSER_DESC->buffer, 0, FUSER_DESC->dat_size);
        FUSER_DESC->dat_size = 0;
    }

    /* change device mode for this process */
    switch (cmd) {
        case DESCVD_IOC_NOOP:
            FUSER_DESC->vd_mode = NOOP;

            /* switching to NOOP zero'es 'user_desc' structure */
            vfree(FUSER_DESC->buffer);
            FUSER_DESC->buffer = NULL;

            FUSER_DESC->vd_mode = 0;
            FUSER_DESC->dat_size = 0;
            FUSER_DESC->buf_size = 0;

            memset(FUSER_DESC->key, 0, sizeof(FUSER_DESC->key));

            printk(KERN_INFO DEVICE_NAME ": mode switched to NOOP.\n");
            break;

        case DESCVD_IOC_REGKEY:
            FUSER_DESC->vd_mode = REGKEY;
            printk(KERN_INFO DEVICE_NAME ": mode switched to REGKEY.\n");
            break;

        case DESCVD_IOC_ENCRYPT:
            FUSER_DESC->vd_mode = ENCRYPT;
            printk(KERN_INFO DEVICE_NAME ": mode switched to ENCRYPT.\n");
            break;

        case DESCVD_IOC_DECRYPT:
            FUSER_DESC->vd_mode = DECRYPT;
            printk(KERN_INFO DEVICE_NAME ": mode switched to DECRYPT.\n");
            break;

        default:    /* this should not ever happen, but just in case... */
            mutex_unlock(&(FUSER_DESC->mutex));
            return -ENOTTY;
    }

    mutex_unlock(&(FUSER_DESC->mutex));
    return 0;
}

